from __future__ import annotations
from dataclasses import dataclass, field
from typing import Callable

from filecheck.options import Options

OP_KINDS = ("DAG", "COUNT", "NOT", "EMPTY", "NEXT", "SAME", "LABEL", "CHECK")


@dataclass(slots=True)
class CheckOp:
    """
    Represents a concrete check instruction (e.g. CHECK-NEXT)
    """

    name: str
    arg: str
    source_line: int
    uops: list[UOp]
    is_literal: bool = field(default=False, kw_only=True)

    def check_line_repr(self, prefix: str = "CHECK"):
        return f"{prefix}{self._suffix()}: {self.arg}"

    def print_source_repr(self, opts: Options):
        print(f"Check rule at {opts.match_filename}:{self.source_line}")
        print(self.check_line_repr(opts.check_prefix))

    def _suffix(self):
        suffix = "{LITERAL}" if self.is_literal else ""
        if self.name == "CHECK":
            return suffix
        return "-" + self.name + suffix


@dataclass(slots=True)
class CountOp(CheckOp):
    """
    special case for COUNT-<number>
    """

    count: int = field(kw_only=True)

    def _suffix(self):
        suffix = "{LITERAL}" if self.is_literal else ""
        return f"-COUNT{self.count}" + suffix


@dataclass(frozen=True, slots=True)
class UOp:
    """
    micro-ops, thse make up the filecheck matching logic
    """

    pass


@dataclass(frozen=True, slots=True)
class Literal(UOp):
    """
    literal match
    """

    content: str


@dataclass(frozen=True, slots=True)
class RE(UOp):
    """
    Regular expression matching
    """

    content: str


@dataclass(frozen=True, slots=True)
class Capture(UOp):
    """
    Variable capture expression

    Stuff like this:
    ```
    ; CHECK: test5:
    ; CHECK:    notw     [[REGISTER:%[a-z]+]]
    ; CHECK:    andw     {{.*}}[[REGISTER]]
    ```
    """

    name: str
    pattern: str
    value_mapper: Callable[[str], int] | Callable[[str], str]


@dataclass(frozen=True, slots=True)
class Subst(UOp):
    """
    Variable substitution, e.g. expect the contents of a variable here.

    Stuff like this:
    ```
    ; CHECK: test5:
    ; CHECK:    notw     [[REGISTER:%[a-z]+]]
    ; CHECK:    andw     {{.*}}[[REGISTER]]
    ````
    """

    variable: str


@dataclass(frozen=True, slots=True)
class NumSubst(UOp):
    """
    Numeric substitution (substitute variable with derived expression).

    Stuff like this:
    ```
    ; CHECK: load r[[#REG:]], [r0]
    ; CHECK: load r[[#REG+1]], [r1]
    ; CHECK: Loading from 0x[[#%x,ADDR:]]
    ; CHECK-SAME: to 0x[[#ADDR + 7]]
    ```
    """

    variable: str
    expr: str



"""
Manages the file input and the position we are at
"""

from __future__ import annotations

import re
import sys
from dataclasses import dataclass, field

from filecheck.options import Options


@dataclass
class FInput:
    """
    A wrapper around file input.

    Handles position keeping and regex searching.
    """

    fname: str
    content: str
    pos: int = field(default=0)

    line_no: int = field(default=0)

    @staticmethod
    def from_opts(opts: Options) -> FInput:
        """
        Create a FInput object from options objects
        """
        # treat - as stding
        if opts.input_file == "-":
            f = sys.stdin
        else:
            f = open(opts.input_file, "r")
        return FInput(opts.input_file, f.read())

    def advance_by(self, dist: int):
        """
        Move forward by dist characters in the input
        """
        assert dist >= 0
        self.line_no += self.content.count("\n", self.pos, self.pos + dist)
        self.pos += dist

    def move_to(self, new_pos: int):
        """
        Move forwards or backwards to a specific point
        """
        sign = 1 if new_pos > self.pos else -1
        lines = self.content.count("\n", min(new_pos, self.pos), max(new_pos, self.pos))
        self.line_no += sign * lines
        self.pos = new_pos

    def match(self, pattern: re.Pattern[str]) -> re.Match[str] | None:
        """
        Match (exactly from the current position)
        """
        return pattern.match(self.content, pos=self.pos)

    def find(
        self, pattern: re.Pattern[str], this_line: bool = False
    ) -> re.Match[str] | None:
        """
        Find the first occurance of a pattern, might be far away.

        If this_line is given, match only until the next newline.
        """
        endpos = self.content.find("\n", self.pos) if this_line else -1
        if endpos == -1:
            endpos = sys.maxsize
        return pattern.search(self.content, pos=self.pos, endpos=endpos)

    def find_between(
        self, pattern: re.Pattern[str], start: int, end: int
    ) -> re.Match[str] | None:
        """
        Find the first occurance of a pattern, might be far away.
        """
        return pattern.search(self.content, pos=start, endpos=end)

    def print_line_with_current_pos(self, pos_override: int | None = None):
        """
        Print the current position in the input file.
        """
        fname = self.fname if self.fname != "-" else "stdin"
        pos = self.pos if pos_override is None else pos_override
        next_newline_at = self.content.find("\n", pos)

        # print the next line if we are pointing at a line end.
        if next_newline_at == pos:
            pos += 1
            next_newline_at = self.content.find("\n", pos)

        last_newline_at = self.start_of_line(pos)
        char_pos = pos - last_newline_at
        print(f"Matching at {fname}:{self.line_no}:{char_pos}")
        print(self.content[last_newline_at + 1 : next_newline_at])
        print(" " * (char_pos - 1), end="^\n")

    def start_of_line(self, pos: int | None = None) -> int:
        """
        Find the start of the line at position pos (defaults to current position)
        """
        if pos is None:
            pos = self.pos
        return max(self.content.rfind("\n", 0, pos), 0)

    def skip_to_end_of_line(self):
        """
        Move to the next \n token (might be at cursor already, then it's a nop)
        """
        if self.pos == 0:
            return
        next_newline = self.content.find("\n", self.pos)
        self.move_to(next_newline)

    def is_end_of_line(self) -> bool:
        """
        Check if line ending or EOF has been reached
        """
        # line ending check
        if self.content.startswith("\n", self.pos):
            return True
        # eof check
        if self.pos == len(self.content) - 1:
            return True
        return False



import re
from typing import Callable

POSIX_REGEXP_PATTERN = re.compile(
    r"\[:(alpha|upper|lower|digit|alnum|xdigit|space|blank|print|punct|graph|word|ascii|cntrl):]"
)
POSIX_REGEXP_REPLACEMENTS = {
    "alpha": "[A-Za-z]",
    "upper": "[A-Z]",
    "lower": "[a-z]",
    "digit": "[0-9]",
    "alnum": "[A-Za-z0-9]",
    "xdigit": "[A-Fa-f0-9]",
    "space": r"\s",
    "blank": r"[ \t]",
    "word": r"\w+",
}


def posix_to_python_regex(expr: str) -> str:
    """
    We need to translate things like `[:alpha:]` to `[A-Za-z]`, etc.

    LLVM supports them, but pythons regex doesn't.
    """
    while (match := POSIX_REGEXP_PATTERN.search(expr)) is not None:
        if match.group(1) not in POSIX_REGEXP_REPLACEMENTS:
            raise ValueError(
                f"Can't translate posix regex, unknown character set: {match.group(1)}"
            )
        expr = expr.replace(match.group(0), POSIX_REGEXP_REPLACEMENTS[match.group(1)])
    return expr


ENCODINGS_MAP = {
    "u": r"\d",
    "d": r"[+-]?\d",
    "x": "[a-f0-9]",
    "X": "[A-F0-9]",
}


def pattern_from_num_subst_spec(
    digits: str | None, encoding: str | None
) -> tuple[str, Callable[[str], int]]:
    digits_expr = "+" if digits is None else f"{{{int(digits[1:])}}}"
    if encoding is None:
        encoding = "u"
    return f"{ENCODINGS_MAP[encoding]}{digits_expr}", (
        hex_int if encoding.lower() == "x" else int
    )


def hex_int(v: str):
    return int(v, base=16)



HELP_TEXT = """filecheck - A Python-native clone of LLVMs FileCheck tool
usage: filecheck FLAGS check-file

FLAGS:
--input-file <file>             : Specify an input file, default is stdin (-)
--check-prefix <prefix>         : Check line prefix (default is CHECK)
--strict-whitespace             : Don't ignore indentation
--comment-prefixes <a>,<b>,<c>  : Ignore lines starting with these prefixes, even if
                                  they contain check lines (default RUN,COM)
--enable-var-scope              : Enables scope for regex variables. Variables not
                                  starting with $ are deleted after each CHECK-LABEL
                                  match.
--match-full-lines              : Expect every check line to match the whole line.


ARGUMENTS:
check-file                      : The file from which the check lines are to be read

This tries to be as close a clone of LLVMs FileCheck as possible. We use it in xDSL
for our tests. If a feature is missing, or behaviour deviates from LLVMs FileCheck,
please file a bug at github.com/AntonLydike/filecheck.
"""



import re

from filecheck.error import ParseError
from filecheck.ops import Literal, RE, Capture, NumSubst, Subst, CheckOp
from filecheck.options import Options

UNESCAPED_BRACKETS = re.compile(r"^\(|[^\\]\(")

CHECK_EMPTY_EXPR = re.compile(r"[^\n]*\n\n")


def compile_uops(
    check: CheckOp, variables: dict[str, str | int], opts: Options
) -> tuple[re.Pattern[str], dict[Capture, int]]:
    """
    Compile a series of uops, given a set of variables, to a regex pattern and an
    extraction dictionary.

    The extraction dictionary tells you which Capture can be found in which regex
    group.
    """
    groups = 0
    expr: list[str] = []
    if check.name == "NEXT":
        expr.append(r"\n?[^\n]*")
    elif check.name == "EMPTY":
        return CHECK_EMPTY_EXPR, dict()

    captures: dict[Capture, int] = dict()
    var_to_group: dict[str, int] = dict()

    for uop in check.uops:
        if isinstance(uop, Literal):
            # literals are matched as is
            if opts.strict_whitespace:
                expr.append(re.escape(uop.content))
            else:
                # TODO: fix this mess
                # basically, I need to replace all whitespaces in the original literal by "\s+"
                # but I still want to regex escape them
                # and re.sub doesn't let me insert \s into the new string for some reason......
                expr.append(
                    re.sub(r"(\\ )+", " ", re.escape(uop.content)).replace(" ", r"\s+")
                )

        elif isinstance(uop, RE):
            # For regexes, we must make sure that we count the number of capture groups
            # present, so that we know which ones contain the Captures, and which ones
            # contain "user supplied" groups.

            # count number of capture groups by counting unescaped brackets
            groups += len(UNESCAPED_BRACKETS.findall(uop.content))
            expr.append(uop.content)
        elif isinstance(uop, Capture):
            # record the group we capture in the dictionary
            captures[uop] = groups + 1
            # this is used to match same-line use of variables
            # like this:
            # // CHECK: alloc [[REG:[a-z]+]], [[REG]]
            # here we can't insert the value now, as it will only be determined when we
            # match the pattern. Luckily, python regex has backwards references.
            var_to_group[uop.name] = groups + 1
            expr.append(f"({uop.pattern})")
        elif isinstance(uop, Subst):
            # if we have substitutions, check if the variable is defined in this line
            if uop.variable in var_to_group:
                expr.append(f"\\{var_to_group[uop.variable]}")
            else:
                # otherwise match immediate
                if uop.variable not in variables:
                    raise ParseError(
                        f"Variable {uop.variable} referenced before assignment",
                        check.source_line,
                        0,
                        check.check_line_repr(),
                    )
                expr.append(re.escape(str(variables[uop.variable])))
        elif isinstance(uop, NumSubst):
            # we don't do numerical substitutions yet
            raise NotImplementedError("Numerical substitutions not supported!")

    # make sure CHECK-LABEL consumes entire line
    if check.name == "LABEL":
        expr.append(r"[^\n]*")

    return re.compile("".join(expr)), captures



from dataclasses import dataclass, field
from typing import Iterator, Callable

from filecheck.finput import FInput
from filecheck.ops import CheckOp, CountOp
from filecheck.options import Options
from filecheck.parser import Parser
from filecheck.compiler import compile_uops
from filecheck.error import CheckError, ParseError


@dataclass
class Context:
    """
    A context object, carrying life variables and such
    """

    live_variables: dict[str, str | int] = field(default_factory=dict)

    negative_matches_stack: list[CheckOp] = field(default_factory=list)
    """
    Keep a stack of CHECK-NOTs around, as we only know the range on which to match
    once we hit the next non-negative check-line.
    """

    negative_matches_start: int | None = field(default=None)


@dataclass
class Matcher:
    """
    The matcher contains the logic for applying the matching, recording variables and
    advancing the file position accordingly.
    """

    opts: Options
    file: FInput
    operations: Parser

    ctx: Context = field(default_factory=Context)

    @classmethod
    def from_opts(cls, opts: Options):
        """
        Construct a matcher form an options object.
        """
        ops = Parser.from_opts(opts)
        fin = FInput.from_opts(opts)
        return Matcher(opts, fin, ops)

    def run(self) -> int:
        """
        Run the matching, returns the exit code of the program.

        Prints a nice message when it fails.
        """
        function_table: dict[str, Callable[[CheckOp], None]] = {
            "DAG": self.check_dag,
            "COUNT": self.check_count,
            "NOT": self.enqueue_not,
            "EMPTY": self.check_empty,
            "NEXT": self.match_immediately,
            "SAME": self.match_eventually,
            "LABEL": self.check_label,
            "CHECK": self.match_eventually,
        }
        checks = 0
        op: CheckOp | None = None
        try:
            for op in self.operations:
                checks += 1
                self._pre_check(op)
                function_table.get(op.name, self.fail_op)(op)
                self._post_check(op)

            # run the post-check one last time to make sure all NOT checks are taken care of.
            self.file.pos = len(self.file.content) - 1
            self._post_check(CheckOp("NOP", "", -1, []))
        except CheckError as ex:
            print(
                f"{self.opts.match_filename}:{self.operations.line_no}: error: {ex.message}"
            )
            self.file.print_line_with_current_pos()

            if ex.pattern:
                print(f"Trying to match with regex '{ex.pattern.pattern}'")
                if match := self.file.find(ex.pattern):
                    print("Possible match at:")
                    self.file.print_line_with_current_pos(match.start(0))

            return 1
        except ParseError as ex:
            print(f"{self.opts.match_filename}:{ex.line_no}:{ex.offset} {ex.message}")
            print(ex.offending_line.rstrip("\n"))
            print(" " * (ex.offset - 1) + "^")
            return 1

        if checks == 0:
            print(
                f"Error: No check strings found with prefix {self.opts.check_prefix}:"
            )
            return 2
        return 0

    def _pre_check(self, op: CheckOp):
        if op.name == "NEXT":
            self.file.skip_to_end_of_line()

    def _post_check(self, op: CheckOp):
        if op.name != "NOT":
            # work through CHECK-NOT checks
            for check in self.ctx.negative_matches_stack:
                self.check_not(check, self.ctx.negative_matches_start)
            # reset the state
            if self.ctx.negative_matches_stack:
                self.ctx.negative_matches_start = None
                self.ctx.negative_matches_stack = []
        elif self.opts.match_full_lines:
            if not self.file.is_end_of_line():
                raise CheckError("Didn't match whole line")

    def check_dag(self, op: CheckOp) -> None:
        raise NotImplementedError()

    def check_count(self, op: CheckOp) -> None:
        # invariant preserved by parser
        assert isinstance(op, CountOp)
        for _ in range(op.count):
            self.match_eventually(op)

    def check_not(self, op: CheckOp, start: int | None):
        """
        Check that op doesn't match between start and current position.
        """
        pattern, _ = compile_uops(op, self.ctx.live_variables, self.opts)
        if start is None:
            start = 0
        if self.file.find_between(pattern, start, self.file.pos):
            raise CheckError(
                f"{self.opts.check_prefix}-NOT: excluded string found in input ('{op.arg}')"
            )

    def enqueue_not(self, op: CheckOp):
        """
        Enqueue a CHECK-NOT operation to be checked at a later time.

        Since CHECK-NOT matches between the last matched line, and the next matched
        line, we need to postpone matching until we know when the next checked line is.

        """
        if self.ctx.negative_matches_start is None:
            self.ctx.negative_matches_start = self.file.pos
        self.ctx.negative_matches_stack.append(op)

    def check_label(self, op: CheckOp):
        """
        https://llvm.org/docs/CommandGuide/FileCheck.html#the-check-label-directive

        Looks for a uniquely identified line in the source file.

        CHECK-LABEL: directives cannot contain variable definitions or uses.
        """
        pattern, _ = compile_uops(op, self.ctx.live_variables, self.opts)
        # match in whole file
        matches = tuple(pattern.finditer(self.file.content))
        # check that we found exactly one match
        if not matches:
            raise CheckError(f'Couldn\'t match "{op.arg}".', pattern=pattern)
        if len(matches) != 1:
            raise CheckError(f"Non-unique {op.check_line_repr()} found")
        # move to match, if it's not already matched.
        (match,) = matches
        new_pos = match.end(0)
        if new_pos < self.file.pos:
            raise CheckError("Label was already checked")

        self.file.move_to(new_pos)

        # remove non $ variables if option is set
        if self.opts.enable_var_scope:
            self.purge_variables()

    def check_empty(self, op: CheckOp):
        # check immediately
        self.match_immediately(op)
        # roll back the last newline, so that we are located at the end of the last line
        # instead of at the start of the next line. This is important for CHECK-NEXT
        # and friends
        self.file.move_to(self.file.pos - 1)

    def match_immediately(self, op: CheckOp):
        """
        Check that the ops pattern is matched immediately.

        Uses file.match
        """
        pattern, repl = compile_uops(op, self.ctx.live_variables, self.opts)
        if match := self.file.match(pattern):
            self.file.move_to(match.end(0))
            for var, group in repl.items():
                self.ctx.live_variables[var.name] = var.value_mapper(match.group(group))
        else:
            raise CheckError(f'Couldn\'t match "{op.arg}".', pattern=pattern)

    def match_eventually(self, op: CheckOp):
        """
        Check that the ops pattern is matched eventually.

        Uses file.find
        """
        pattern, repl = compile_uops(op, self.ctx.live_variables, self.opts)
        if match := self.file.find(pattern, op.name == "SAME"):
            self.file.move_to(match.end())
            for var, group in repl.items():
                self.ctx.live_variables[var.name] = var.value_mapper(match.group(group))

        else:
            raise CheckError(f'Couldn\'t match "{op.arg}".', pattern=pattern)

    def fail_op(self, op: CheckOp):
        """
        Helper used for unknown operations. Should not occur as we check when parsing.
        """
        raise RuntimeError(f"Unknown operation: {op}")

    def purge_variables(self):
        """
        Purge variables not starting with '$'
        """
        for key in tuple(self.ctx.live_variables.keys()):
            if not key.startswith("$"):
                self.ctx.live_variables.pop(key)



from dataclasses import dataclass
from typing import Iterable


@dataclass
class Options:
    match_filename: str
    input_file: str = "-"
    check_prefix: str = "CHECK"
    strict_whitespace: bool = False
    enable_var_scope: bool = False
    match_full_lines: bool = False
    comment_prefixes: list[str] = "COM,RUN"  # type: ignore[reportAssignmentType]

    def __post_init__(self):
        # make sure we split the comment prefixes
        if isinstance(self.comment_prefixes, str):
            self.comment_prefixes = self.comment_prefixes.split(",")


def parse_argv_options(argv: list[str]) -> Options:
    # pop the name off of argv
    _ = argv.pop(0)

    # final options to return
    opts = {}
    # args that were consumed
    remove: list[int] = []
    argv = list(normalise_args(argv))

    for i, arg in enumerate(argv):
        if arg.startswith("--"):
            arg = arg[2:].replace("-", "_")
        elif arg.startswith("-"):
            arg = arg[1:].replace("-", "_")
        else:
            continue
        if arg not in Options.__dataclass_fields__:
            continue

        remove.append(i)
        if Options.__dataclass_fields__[arg].type == bool:
            opts[arg] = True
        else:
            if i == len(argv) - 1:
                raise RuntimeError("Out of range arguments")
            opts[arg] = argv[i + 1]
            remove.append(i + 1)

    for idx in sorted(remove, reverse=True):
        argv.pop(idx)

    if len(argv) > 1:
        raise RuntimeError(
            f"Unconsumed arguments: {argv}, expected one remaining arg, the match-filename."
        )

    opts["match_filename"] = argv[0]

    return Options(**opts)  # pyright: ignore[reportUnknownArgumentType]


def normalise_args(argv: list[str]) -> Iterable[str]:
    """
    Normalize arguments by splitting "--key=value" pairs into separate "--key" and
    "value" entries.
    """
    for arg in argv:
        if arg.startswith("-") and "=" in arg:
            yield from arg.split("=", 1)
        else:
            yield arg



import re
from typing import Any


class CheckError(Exception):
    message: str
    pattern: re.Pattern[str] | None

    def __init__(self, msg: str, *args: Any, pattern: re.Pattern[str] | None = None):
        self.message = msg
        self.pattern = pattern
        super().__init__(args)


class ParseError(Exception):
    message: str
    line_no: int
    offset: int
    offending_line: str

    def __init__(
        self, message: str, line_no: int, offset: int, offending_line: str, *args: Any
    ):
        super().__init__(*args)
        self.message = message
        self.line_no = line_no
        self.offset = offset
        self.offending_line = offending_line



"""
parser for filecheck syntax
"""

from dataclasses import dataclass, field
from typing import Iterator, TextIO
import re

from filecheck.error import ParseError
from filecheck.ops import CheckOp, Literal, RE, Capture, Subst, NumSubst, UOp, CountOp
from filecheck.options import Options, parse_argv_options
from filecheck.regex import posix_to_python_regex, pattern_from_num_subst_spec


def pattern_for_opts(opts: Options) -> tuple[re.Pattern[str], re.Pattern[str]]:
    return re.compile(
        re.escape(opts.check_prefix)
        + r"(-(DAG|COUNT-\d+|NOT|EMPTY|NEXT|SAME|LABEL))?(\{LITERAL})?:\s?([^\n]*)\n?"
    ), re.compile(
        "("
        + "|".join(map(re.escape, opts.comment_prefixes))
        + ").*"
        + re.escape(opts.check_prefix)
    )


# see https://llvm.org/docs/CommandGuide/FileCheck.html#filecheck-string-substitution-blocks
VAR_CAPTURE_PATTERN = re.compile(r"\[\[(\$?[a-zA-Z_][a-zA-Z0-9_]*):([^\n]+)]]")

# see https://llvm.org/docs/CommandGuide/FileCheck.html#filecheck-string-substitution-blocks
VAR_SUBST_PATTERN = re.compile(r"\[\[(\$?[a-zA-Z_][a-zA-Z0-9_]*)]]")

# numeric substitution blocks, see:
# https://llvm.org/docs/CommandGuide/FileCheck.html#filecheck-numeric-substitution-blocks
NUMERIC_CAPTURE_PATTERN = re.compile(
    r"\[\[#(%(\.[0-9]+)?([udxX])?,)?((\$?[a-zA-Z_][a-zA-Z0-9_]+):(\d+)?)?]]"
)

# numeric substitution blocks, see:
# https://llvm.org/docs/CommandGuide/FileCheck.html#filecheck-numeric-substitution-blocks
NUMERIC_SUBST_PATTERN = re.compile(
    r"\[\[#(\$?[a-zA-Z_][a-zA-Z0-9_]*)([a-z0-9 +\-()]*)]]"
)


LINE_SPLIT_RE = split = re.compile(r"(\{\{|\[\[|]]|}})")


@dataclass
class Parser(Iterator[CheckOp]):
    """
    parse a file containing filecheck directives and returns all found check directives
    """

    opts: Options

    input: TextIO

    check_line_regexp: re.Pattern[str]
    comment_line_regexp: re.Pattern[str]

    line_no: int = field(default=0)
    line: str = field(default=0)

    @classmethod
    def from_opts(cls, opts: Options):
        return Parser(opts, open(opts.match_filename), *pattern_for_opts(opts))

    def __next__(self) -> CheckOp:
        """
        Scans forward in self.input until a line matching  self.check_line_regexp is
        found. Returns the kind (CHECK/NEXT/DAG/EMPTY/NOT) and the arg (whatever comes
        after the colon).

        Note that for all CHECK-<thing>: it returns <thing> as kind, but for CHECK: it
        returns CHECK.
        """
        while True:
            self.line_no += 1
            line = self.input.readline()
            if line == "":
                raise StopIteration()
            # skip any lines containing comment markers before checks
            if self.comment_line_regexp.search(line) is not None:
                continue
            match = self.check_line_regexp.search(line)
            # no check line = skip
            if match is None:
                continue
            kind = match.group(2)
            literal = match.group(3)
            arg = match.group(4)
            if kind is None:
                kind = "CHECK"
            if arg is None:
                arg = ""
            # verify that non-empty checks have an actual thing to match
            if kind != "EMPTY":
                if not arg:
                    raise ParseError(
                        f"found empty check string with prefix '{kind}:'",
                        self.line_no,
                        match.start(3),
                        line,
                    )
            if not self.opts.strict_whitespace:
                arg = arg.strip()

            # parse the uops, but only if we are not in LITERAL mode
            uops: list[UOp]
            if literal is None:
                uops = self.parse_args(arg, line)
            else:
                uops = [Literal(arg)]

            # special case for COUNT ops
            if kind.startswith("COUNT"):
                count = int(kind[6:])
                if count == 0:
                    raise ParseError(
                        f"invalid count in -COUNT specification on prefix '{opts.check_prefix}' (count can't be 0)",
                        self.line_no,
                        match.end(2),
                        line,
                    )
                return CountOp(
                    "COUNT",
                    arg,
                    self.line_no,
                    uops,
                    is_literal=literal is not None,
                    count=count,
                )
            return CheckOp(
                kind, arg, self.line_no, uops, is_literal=literal is not None
            )

    def parse_args(self, arg: str, line: str) -> list[UOp]:
        """
        parse check args into uops that can later be compiled into a regex

        Returns a list of uops
        """
        uops: list[UOp] = []
        parts = LINE_SPLIT_RE.split(arg)
        offset = len(line) - len(arg)
        while parts:
            part = parts.pop(0)
            if part == "[[":
                # grab parts greedily until we hit a ]]
                while not part.endswith("]]"):
                    if not parts:
                        raise ParseError(
                            "Invalid substitution block, no ]]",
                            self.line_no,
                            offset,
                            line,
                        )
                    part += parts.pop(0)
                # check if we are a simple capture pattern [[<name>:<regex>]]
                if match := VAR_CAPTURE_PATTERN.fullmatch(part):
                    uops.append(
                        Capture(
                            match.group(1), posix_to_python_regex(match.group(2)), str
                        )
                    )
                # check if we are a simple substitution pattern: [[<<name>>]]
                elif match := VAR_SUBST_PATTERN.fullmatch(part):
                    uops.append(Subst(match.group(1)))
                # check if we are a numeric substitution pattern [[#<name>(<expr>)?]]
                elif match := NUMERIC_SUBST_PATTERN.fullmatch(part):
                    # simplify to non-numeric substitution if expr is empty
                    if match.group(2) is None or match.group(2).strip() == "":
                        uops.append(Subst(match.group(1)))
                    else:
                        uops.append(NumSubst(match.group(1), match.group(2)))
                # numeric capture patterns are a tricky bunch
                # see https://llvm.org/docs/CommandGuide/FileCheck.html#filecheck-numeric-substitution-blocks
                elif match := NUMERIC_CAPTURE_PATTERN.fullmatch(part):
                    pattern, mapper = pattern_from_num_subst_spec(
                        match.group(2), match.group(3)
                    )
                    uops.append(Capture(match.group(5), pattern, mapper))
                else:
                    raise ParseError(
                        f"Invalid substitution block, unknown format: {part}",
                        self.line_no,
                        offset,
                        line,
                    )
            elif part == "{{":
                while not part.endswith("}}"):
                    if not parts:
                        raise ParseError(
                            "Invalid regex block, no }}", self.line_no, offset, line
                        )
                    part += parts.pop(0)

                pattern = part[2:-2]
                uops.append(RE(posix_to_python_regex(pattern)))
            elif part != "":
                uops.append(Literal(part))
            offset += len(part)
        return uops


if __name__ == "__main__":
    import sys

    opts = parse_argv_options(sys.argv)
    for p in Parser.from_opts(opts):
        print(p)



import sys

from filecheck.help import HELP_TEXT
from filecheck.matcher import Matcher
from filecheck.options import parse_argv_options


def main(argv: list[str] | None = None):
    if argv is None:
        argv = sys.argv

    if "--help" in argv or len(argv) < 2:
        print(HELP_TEXT)
        return

    opts = parse_argv_options(argv)
    m = Matcher.from_opts(opts)
    sys.exit(m.run())
